local render = {}
local utf8 = require("utf8")

local function dashedLine(ox,oy,tx,ty,pat,type)
    local dist = math.sqrt((tx-ox)^2 + (ty-oy)^2)
    local dx,dy = tx-ox,ty-oy
    local loop = pat[1] + pat[2]
    for i = 0,dist,loop do
        k0 = i/dist
        k1 = (i+pat[1])/dist
        love.graphics.line(ox+k0*dx,oy+k0*dy,ox+k1*dx,oy+k1*dy)
    end
end

local function split(str,keyword)
	local tab={}
	local index=1
	local from=1
	local to=1
	while true do
		if string.sub(str,index,index)==keyword then
			to=index-1
			if from>to then 
				table.insert(tab, "")
			else
				table.insert(tab, string.sub(str,from,to))
			end
			from=index+1
		end
		index=index+1
		if index>string.len(str) then
			if from<=string.len(str) then
				table.insert(tab, string.sub(str,from,string.len(str)))
			end
			return tab
		end
	end
end

local function copy(source,copyto,ifcopyfunction)
	copyto=copyto or {}
	for k, v in pairs(source or {}) do
		if type(v) == "table" then
			copyto[k] = table.copy(v,copyto[k])          
		elseif type(v) == "function" then 
			if ifcopyfunction then
				copyto[k] = v
			end
		else 
			copyto[k] = v
		end
	end
	return copyto
end

local function dividLines(text,font,lineLimit)
    local lines = {}
    local textLen = utf8.len(text)
    local u = 0
    local line = ""
    for i = 1,textLen do
        local char = string.sub(text,utf8.offset(text,i),utf8.offset(text,i+1)-1)
        local cw = font:getWidth(char)
        u = u + cw
        if u> lineLimit then
            table.insert(lines,line)
            u = cw
            line = char
        else
            line = line .. char
        end
            
    end
    table.insert(lines,line)
    return lines
end

local function getAlinePos(aline,w,h,tw,th,line,spacing,lineCount)
    line = line or 1
    spacing = spacing or 0
    lineCount = lineCount or 1
    local x,y
    if aline.x == "left" then
        x = 0
    elseif aline.x == "center" then
        x = w/2 - tw/2
    elseif aline.x == "right" then
        x = w - tw
    end
    if aline.y == "top" then
        y = (line - 1) * (th + spacing) + spacing
    elseif aline.y == "center" then
        y = h/2 - ((th + spacing) * lineCount - spacing )/2 + (th + spacing)* (line - 1)
    elseif aline.y == "bottom" then
        y = h - (th + spacing) * lineCount - spacing + (th + spacing)* (line - 1)
    end
    return x,y
end

local function limitInMask(style,opt,drawFunc,pass)
    if style.noStencil then return drawFunc() end
    if pass then
        local mask = function()
            if opt.container then
                love.graphics.rectangle("fill",unpack(opt.container)) 
            end
        end
        love.graphics.stencil(mask,"increment")
        love.graphics.setStencilTest("equal",opt.container and 1 or 0)
        drawFunc()
        love.graphics.setStencilTest()
    else
        local x,y,w,h = opt.x,opt.y,opt.w,opt.h
        local state = "normal"
        local sx,sy = style.scale[state].x,style.scale[state].y
        local offx,offy = style.offset[state].x,style.offset[state].y
        if opt.noOffset then offx = 0;offy = 0 end
        if opt.noScale then sx = 1;sy = 1 end
        local xx,yy,ww,hh = x+offx,y+offy,w*sx,h*sy
        local rx,ry = style.roundCorner.rx,style.roundCorner.ry
        local fillColor = style.fillColor[state]
        if opt.noRound then rx,ry = 0,0 end
        local mask = function()
            if opt.container then
                love.graphics.rectangle("fill",unpack(opt.container)) 
            end
            
            if opt.mask then
                love.graphics.rectangle("fill",unpack(opt.mask)) 
            else
                love.graphics.rectangle("fill",xx,yy,ww,hh,rx,ry)  
            end
            
        end
        love.graphics.stencil(mask,"increment")
        love.graphics.setStencilTest("equal",opt.container and 2 or 1)
        drawFunc()
        love.graphics.setStencilTest()
    end
end

local function renderLine(style,ox,oy,tx,ty)
    local state = "normal"
    local strokeColor = style.strokeColor[state]
    local strokeWidth = style.strokeWidth[state]
    love.graphics.setLineWidth(strokeWidth)
    love.graphics.setPointSize(strokeWidth)
    love.graphics.setColor(strokeColor)
    if style.strokeDash.enable then
        local pat = style.strokeDash.pattern
        local type = style.strokeDash.type
        dashedLine(ox,oy,tx,ty,pat,type)
    else
        love.graphics.line(ox,oy,tx,ty)
    end
end

local function renderText(style,opt)
    if opt.noText then return end
    local x,y,w,h = opt.x,opt.y,opt.w,opt.h
    local state = not opt.state and "normal" or opt.state
    local sx,sy = style.scale[state].x,style.scale[state].y
    local offx,offy = style.offset[state].x,style.offset[state].y
    if opt.noOffset then offx = 0;offy = 0 end
    if opt.noScale then sx = 1;sy = 1 end
    local font = opt.font[state].obj
    local text = opt.text
    local color = style.fontColor[state]
    local xx,yy,ww,hh = x+offx+2,y+offy+2,w*sx-4,h*sy-4
    local cx,cy = xx+ww/2,yy + hh/2
    local aline = opt.aline or {x = "center",y = "center"}
    local fw,fh = font:getWidth("W"),font:getHeight("H")
    
    limitInMask(style,opt,function()
        love.graphics.setColor(color)
        love.graphics.setFont(font)
        local lineSpacing = opt.lineSpacing or 0
        if string.find(text,"\n") then --自动多行
            local lines = opt.lines or split(text,"\n")
            opt.lines = lines
            local lineCount = #lines
            for i = 1, lineCount do
                local line = lines[i]
                local lw = font:getWidth(line)
                local lx,ly = getAlinePos(aline,ww,hh,lw,fh,i,lineSpacing,lineCount)
                love.graphics.print(line,math.floor(xx + lx),math.floor(yy + ly))
            end
        elseif opt.multiLine then --自适应多行
            local lines = opt.lines or dividLines(text,font,ww)
            opt.lines = lines
            local lineCount = #lines
            for i = 1, lineCount do
                local line = lines[i]
                local lw = font:getWidth(line)
                local lx,ly = getAlinePos(aline,ww,hh,lw,fh,i,lineSpacing,lineCount)
                love.graphics.print(line,math.floor(xx + lx),math.floor(yy + ly))
            end
        elseif opt.wrapLine then --走马灯
            local lw = font:getWidth(text)
            opt.wrapOffset = opt.wrapOffset or 0
            opt.wrapOffset = opt.wrapOffset +love.timer.getDelta()*(lw+ww)/3
            opt.wrapOffset = opt.wrapOffset > lw+ww and 0 or opt.wrapOffset 
            --opt.wrapOffset = 0
            love.graphics.print(text,math.floor(xx - opt.wrapOffset ),math.floor(yy + hh/2-fh/2))
            love.graphics.print(text,math.floor(xx - opt.wrapOffset + lw+ww ),math.floor(yy + hh/2-fh/2))
        else --单行
            local lw = font:getWidth(text)
            local lx,ly =  getAlinePos(aline,ww,hh,lw,fh,1,lineSpacing,1)
            love.graphics.print(text,math.floor(xx + lx),math.floor(yy + ly))
        end
    end)
end

local function renderRect_fill(style,opt)
    if opt.noFill then return end
    local x,y,w,h = opt.x,opt.y,opt.w,opt.h
    local state = not opt.state and "normal" or opt.state
    local sx,sy = style.scale[state].x,style.scale[state].y
    local offx,offy = style.offset[state].x,style.offset[state].y
    if opt.noOffset then offx = 0;offy = 0 end
    if opt.noScale then sx = 1;sy = 1 end
    local xx,yy,ww,hh = x+offx,y+offy,w*sx,h*sy
    local rx,ry = style.roundCorner.rx,style.roundCorner.ry
    local fillColor = style.fillColor[state]
    if opt.noRound then rx,ry = 0,0 end
    limitInMask(style,opt,function()
        love.graphics.setColor(fillColor)
        love.graphics.rectangle("fill",xx,yy,ww,hh,rx,ry)      
    end)
end

local function renderRect_grad(style,opt)
    if opt.noGrad then return end
    local x,y,w,h = opt.x,opt.y,opt.w,opt.h
    local state = not opt.state and "normal" or opt.state
    local sx,sy = style.scale[state].x,style.scale[state].y
    local offx,offy = style.offset[state].x,style.offset[state].y
    if opt.noOffset then offx = 0;offy = 0 end
    if opt.noScale then sx = 1;sy = 1 end
    local xx,yy,ww,hh = x+offx,y+offy,w*sx,h*sy
    local rx,ry = style.roundCorner.rx,style.roundCorner.ry
    limitInMask(style,opt,function()
        local grad = opt.mesh or style.getMesh() 
        local gradColor = style.gradient.color[state]
        love.graphics.setColor(gradColor)
        love.graphics.draw(grad,xx,yy,0,ww,hh)
    end)
end

local function renderRect_stroke(style,opt)
    if opt.noStroke then return end
    local x,y,w,h = opt.x,opt.y,opt.w,opt.h
    local state = not opt.state and "normal" or opt.state
    local sx,sy = style.scale[state].x,style.scale[state].y
    local offx,offy = style.offset[state].x,style.offset[state].y
    if opt.noOffset then offx = 0;offy = 0 end
    if opt.noScale then sx = 1;sy = 1 end
    local strokeColor = style.strokeColor[state]
    local strokeWidth = style.strokeWidth[state]
    local xx,yy,ww,hh = x+offx,y+offy,w*sx,h*sy
    local rx,ry = style.roundCorner.rx,style.roundCorner.ry
    if opt.noRound then rx,ry = 0,0 end
    if not opt.noStroke then
        limitInMask(style,opt,function()
            love.graphics.setLineWidth(strokeWidth)
            love.graphics.setPointSize(strokeWidth)
            love.graphics.setColor(strokeColor)
            love.graphics.rectangle("line",xx,yy,ww,hh,rx,ry)
        end,true)
    end
end

local function renderRect(style,opt)
    renderRect_fill(style,opt)
    renderRect_grad(style,opt)
    renderRect_stroke(style,opt)
end

local function renderShape(style,opt,shape)
    renderRect_fill(style,opt)
    renderRect_stroke(style,opt)
    
    limitInMask(style,opt,function()
        if opt.toggle ~= false then
            local color = style.strokeColor.active
            love.graphics.setColor(color)
            local ss = opt.w>opt.h and opt.h or opt.w
            local xx,yy = opt.x+opt.w/2-ss/2,opt.y+opt.h/2-ss/2
            if shape == "check" then
                love.graphics.line(xx+0.25*ss,yy+0.25*ss,xx+0.75*ss,yy+0.75*ss)
                love.graphics.line(xx+0.75*ss,yy+0.25*ss,xx+0.25*ss,yy+0.75*ss)
            elseif shape == "radio" then
                love.graphics.circle("fill",xx + 0.5*ss,yy+0.5*ss,0.25*ss)
            elseif shape == "up" then
                love.graphics.polygon("fill",xx + 0.2*ss,yy+0.75*ss,xx+0.5*ss,yy+0.25*ss,xx+0.8*ss,yy+0.75*ss)
            elseif shape == "down" then    
                love.graphics.polygon("fill",xx + 0.2*ss,yy+0.25*ss,xx+0.5*ss,yy+0.75*ss,xx+0.8*ss,yy+0.25*ss)
            elseif shape == "left" then
                love.graphics.polygon("fill",xx + 0.75*ss,yy+0.2*ss,xx+0.25*ss,yy+0.5*ss,xx+0.75*ss,yy+0.8*ss)
            elseif shape == "right" then
                love.graphics.polygon("fill",xx + 0.25*ss,yy+0.2*ss,xx+0.75*ss,yy+0.5*ss,xx+0.25*ss,yy+0.8*ss)
            end
        end
    end)
end

local function renderImage(style,opt)
    local x,y,w,h = opt.x,opt.y,opt.w,opt.h
    local state = not opt.state and "normal" or opt.state
    local sx,sy = style.scale[state].x,style.scale[state].y
    local offx,offy = style.offset[state].x,style.offset[state].y
    if opt.noOffset then offx = 0;offy = 0 end
    if opt.noScale then sx = 1;sy = 1 end
    local strokeColor = style.strokeColor[state]
    local strokeWidth = style.strokeWidth[state]
    local fillColor = style.fillColor[state]
    local xx,yy,ww,hh = x+offx,y+offy,w*sx,h*sy
    local rx,ry = style.roundCorner.rx,style.roundCorner.ry
    local aline = opt.aline or {x = "center",y = "center"}
    local imageColor = style.imageColor[state]
    if opt.noFrame then
        opt.noFill = true
        opt.noStroke = true
    end
    renderRect_fill(style,opt)
    limitInMask(style,opt,function()
        local iw,ih = opt.image:getDimensions()
        love.graphics.setColor(imageColor)
        if opt.adjust then
            love.graphics.draw(opt.image,xx+ww/2,yy+hh/2,opt.rot or 0,ww/iw,hh/ih,iw/2,ih/2)
        else
            local lx,ly = getAlinePos(aline,ww,hh,iw,ih)
            love.graphics.draw(opt.image,xx+lx+iw/2,yy+ly+ih/2,opt.rot or 0,1,1,iw/2,ih/2)
        end
    end)
    renderRect_stroke(style,opt)
end


function render.block(style,opt)
    renderRect(style,opt)
end

function render.button(style,opt)
    renderRect(style,opt)
    renderText(style,opt)
end

function render.shape(style,opt)
    if opt.noDraw then return end
    renderShape(style,opt,opt.shape)
end

function render.label(style,opt)
    renderText(style,opt)
end

function render.image(style,opt)
    renderImage(style,opt)
end

function render.panel(style,opt)
    if opt.withStyle then
        renderRect(style,opt)
    else
        opt.noRound = true
        opt.noScale = true
        opt.noOffset = true
        renderRect_fill(style,opt)
        --renderRect_stroke(style,opt)
    end
end

function render.colorpicker()
    
end

function render.progress(style,opt)
    renderRect_fill(style,opt)
    
    local x,y,w,h = opt.x,opt.y,opt.w,opt.h
    local state = not opt.state and "normal" or opt.state
    local sx,sy = style.scale[state].x,style.scale[state].y
    local offx,offy = style.offset[state].x,style.offset[state].y
    local strokeColor = style.strokeColor[state]
    local strokeWidth = style.strokeWidth[state]
    local fillColor = style.fillColor[state]
    local xx,yy,ww,hh = x+offx,y+offy,w*sx,h*sy
    local rx,ry = style.roundCorner.rx,style.roundCorner.ry
    limitInMask(style,opt,function()
        local grad = opt.mesh or style.getMesh()
        local gradColor = style.gradient.color[state]
        love.graphics.setColor(gradColor)
        local gx,gy = xx,yy
        local gw,gh = ww,hh
        if opt.vert then
            gh = gh * opt.value
            if not opt.counter then
                gy = hh - gh
            end
        else
            gw = gw*opt.value
            if opt.counter then
                gx = ww - gw
            end
        end
        love.graphics.draw(grad,gx,gy,0,gw,gh)
    end)
    renderRect_stroke(style,opt)
    
end

function render.slider(style,opt)
    renderRect_fill(style,opt)
    renderRect_stroke(style,opt)
    if not opt.noValue then
        renderText(style,opt)
    end
end


function render.text(style,opt)
    renderRect(style,opt)
    limitInMask(style,opt,function()
        local g = opt.grid
        if opt.debug then
            love.graphics.setColor(0.2,0.2,1,1)
            for x = 1,g.cols do
                love.graphics.rectangle("line",g.x + (x-1)*g.cw, g.y,g.cw,g.ch)
            end
            love.graphics.setColor(1,0,0,0.3)
            love.graphics.rectangle("fill",g.x,g.y,g.w,g.h)
        end
        local color = style.fontColor.normal
        love.graphics.setColor(color)
        local font = opt.font.normal.obj
        love.graphics.setFont(font)
        for x = 1,g.cols do
            local cell = g.cells[x]
            if cell.t then
                love.graphics.printf(
                    cell.char,math.floor(g.x + (x-1)*g.cw+g.cw/2)-4,math.floor(g.y),
                    g.cw*2,"center")
            end
        end
        
        local c = opt.cursor
        if opt.editing then
            if opt.cursor.mode == "insert" then
                c.timer = c.timer - love.timer.getDelta()
                if c.timer<0 then
                    c.timer = c.cd
                end
                love.graphics.setColor(1,1,1,c.timer/c.cd)
                love.graphics.line(g.x + (c.pos-1)*g.cw,g.y+2,g.x + (c.pos-1)*g.cw,g.y+g.ch-2)
            end
            
        end
    end)
end

function render.tip(style,opt)
    renderRect(style,opt)
    renderText(style,opt)
end

function render.frame(style,opt)
    local x,y,w,h = opt.x,opt.y,opt.w,opt.h
    local state = not opt.state and "normal" or opt.state
    local sx,sy = 1,1
    local offx,offy = 0,0
    local strokeColor = style.strokeColor[state]
    local strokeWidth = style.strokeWidth[state]
    local fillColor = style.fillColor[state]
    local xx,yy,ww,hh = x+offx,y+offy,w*sx,h*sy
    local rx,ry = style.roundCorner.rx,style.roundCorner.ry
    local font = opt.font.normal.obj
    local text = opt.text
    local titleColor = style.fontColor[state]
    local fw,fh = font:getWidth("W"),font:getHeight("H")
    if opt.noRound then
        rx,ry = 0,0
    end
    local title_h = fh+2
    love.graphics.setColor(fillColor)
    love.graphics.rectangle("fill",xx,yy,ww,hh,rx,ry) --out line

    if not opt.noGrad then
        local test = function()
            love.graphics.rectangle("fill",xx,yy,ww,hh,rx,ry)
            love.graphics.rectangle("fill",xx,yy,ww,title_h)
        end
        love.graphics.stencil(test, "increment", 1)
        local grad = opt.mesh or style.getMesh()
        local gradColor = style.gradient.color[state]
        love.graphics.setStencilTest("equal", 2) --title
        love.graphics.setColor(gradColor)
        love.graphics.draw(grad,xx,yy,0,ww,title_h)
        love.graphics.setStencilTest()
    end
    
    
    if not opt.noStroke then
        renderRect_stroke(style,opt)
        if not opt.folded then
            love.graphics.line(xx,yy+title_h,xx+ww,yy+title_h)
        end
    end
    
    if opt.resizable and not opt.folded then
        love.graphics.setColor(strokeColor)
        local rxx = rx<5 and 5 or rx
        local ryy = ry<5 and 5 or ry
        love.graphics.polygon("fill",
            xx+ww-rxx,yy+hh,
            xx+ww,yy+hh,
            xx+ww,yy+hh-ryy)
        love.graphics.setLineWidth(strokeWidth)
        love.graphics.polygon("line",
            xx+ww-rxx,yy+hh,
            xx+ww,yy+hh,
            xx+ww,yy+hh-ryy)
    end
end

function render.canvas(style,opt)
    limitInMask(style,opt,function()
        love.graphics.push()
        love.graphics.translate(opt.x-opt.offsetX,opt.y-opt.offsetY)
        opt.render()
        love.graphics.pop()
    end)
end

function render.tab(style,opt)
    --renderRect(style,opt)
end

return render